

export type Id = [number, number];

export const enum Liveness {
    DEAD = 0,
    ALIVE = 1,
    INACTIVE = 2
}

export interface HasGeneration {
    generation: number;
}
export interface EntityState {
    alive: Liveness;
}

export type Store<T> = (T & HasGeneration)[];
export type SparseStore<T> = Record<number, T & HasGeneration>;
export class Data {
    entity: Store<EntityState> = [];

    constructor(source?: Partial<Data>) {
        if(source?.entity) this.entity = source.entity.slice();
    }
}

export function copy<T>(source: SparseStore<T>): SparseStore<T> {
    return JSON.parse(JSON.stringify(source));
}

// Ergonomic Lookup typings
type StoreKeysOf<DATA> = {
    [K in keyof DATA]: DATA[K] extends Record<number, infer T & HasGeneration> ? K : never;
};
type StoreKeys<DATA extends Data> = StoreKeysOf<DATA>[keyof DATA];
type ItemType<S> = S extends Record<number, infer T & HasGeneration> ? T : never;
type StoreType<DATA extends Data, K> = K extends "id" ? Id : K extends keyof DATA ? ItemType<DATA[K]> : never;
type StoreTypes<DATA extends Data, K extends (keyof DATA | "id")[]> = {
    [I in keyof K]: StoreType<DATA, K[I]>;
};
type MaybeStoreTypes<DATA extends Data, K> = {
    [I in keyof K]: StoreType<DATA, K[I]> | null;
};


/**
 * Create an entity in the store
 * @param data store
 * @param assign map of components to attach
 * @param state Liveness state, allows creating an inactive entity
 * @returns the new entity's ID and generation
 */
type Assigner<DATA extends Data> = {
    [S in StoreKeys<DATA>]?: StoreType<DATA, S>
};
export function Create<DATA extends Data>(data: DATA, assign: Assigner<DATA>, state = Liveness.ALIVE): Id {
    const entities = data.entity;
    // find free ID
    let freeId = -1;
    let generation = -1;
    for(let id = 0; id < entities.length; id++) {
        if(entities[id].alive == Liveness.DEAD) {
            freeId = id;
            generation = entities[id].generation + 1;
            break;
        }
    }

    if(freeId == -1) {
        freeId = entities.length;
        generation = 1;
    }

    entities[freeId] = {
        generation,
        alive: state
    };

    for(const key in assign) {
        const store = data[key as keyof Data] as Store<{}>|SparseStore<{}>;
        store[freeId] = {...(assign as Record<string, HasGeneration>)[key] as {}, generation};
    }

    return [freeId, generation];
}

/**
 * "Delete" an entity
 * @param data store
 * @param id entity ID
 * @param generation entity ID generation
 * @param state can be set to Liveness.INACTIVE to disable an entity without actually killing it, for later resurrection
 */
export function Remove<DATA extends Data>(data: DATA, [id, generation]: Id, state = Liveness.DEAD) {
    if(data.entity[id] && data.entity[id].generation == generation) {
        data.entity[id].alive = state;
    }
}

/**
 * Look up components that may or may not exist for an entity
 * @param data store
 * @param param1 entity Id
 * @param components names of components to look for
 * @returns the cooresponding components, with unfound ones replaced by nulls
 */
export function Lookup<DATA extends Data, K extends StoreKeys<DATA>[]>(data: DATA, [id, generation]: Id, ...components: K): MaybeStoreTypes<DATA, K> {
    const entity = data.entity[id];
    // inactive entities are fine to lookup, but dead ones are not
    if(entity && entity.generation == generation && entity.alive != Liveness.DEAD) {
        return components.map(storeName => {
            const store = data[storeName] as unknown as Store<{}>|SparseStore<{}>;
            const component = store[id];
            if(component && component.generation == generation) {
                return component;
            } else {
                return null;
            }
        }) as MaybeStoreTypes<DATA, K>;
    } else {
        return components.map(() => null) as MaybeStoreTypes<DATA, K>;
    }
}

/**
 * Query a Data collection for all Alive entities possessing the named set of Components.
 * "id" can be used as a pseudo-component to get the Ids for a match
 * @returns an array of tuples containing the matching entity Components
 */
export function Join<DATA extends Data, K extends (StoreKeys<DATA> | "id")[]>(data: DATA, ...components: K): StoreTypes<DATA, K>[] {
    const entities = data.entity;
    const stores = components.map(name => {
        if(name == "id") {
            return "id";
        } else {
            return data[name] as unknown as (Store<{}>|SparseStore<{}>)
        };
    });

    const results: StoreTypes<DATA, K>[] = [];
    const firstStore = stores.filter(store => store !== "id")[0] as Store<{}>|SparseStore<{}>;
    if(Array.isArray(firstStore)) {
        for(let id = 0; id < firstStore.length; id++) {
            JoinLoop(id, entities, stores, results);
        }
    } else {
        for(const id in firstStore) {
            JoinLoop(Number(id), entities, stores, results);
        }
    }
    return results;
}
function JoinLoop<DATA extends Data, K extends (StoreKeys<DATA> | "id")[]>(id: number, entities: Store<EntityState>, stores: (Store<{}> | SparseStore<{}> | "id")[], results: StoreTypes<DATA, K>[]) {
    const fullId: Id = [id, -1];
    const result: (HasGeneration | Id)[] = [];

    let generation = -1;
    for (const store of stores) {
        if(store === "id") {
            result.push(fullId);
            continue;
        }
        const component = store[id];
        if (component && (component.generation == generation || generation == -1)) {
            generation = component.generation;
            result.push(component);
        } else {
            return;
        }
    }

    // only accept active entities (do this check here, where the generation is known)
    const entity = entities[id];
    if (entity.alive != Liveness.ALIVE || generation != entity.generation) return;

    // backpatch generation now that it's known
    fullId[1] = generation;

    results.push(result as StoreTypes<DATA, K>);
}
